
import logging

from django.core.cache import cache
from django.db.models import Manager
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver

logger = logging.getLogger(__name__)


class CacheableMixin:
    CACHE = cache
    CACHED_PROPERTIES = []

    @classmethod
    def cache_key(cls, pk):
        return f"pfx.{cls._meta.db_table}.{pk}"

    @classmethod
    def cache_get(cls, pk):
        return cls.CACHE.get(cls.cache_key(pk))

    def cache(self):
        self.load_cached_properties()
        self.CACHE.set(self.cache_key(self.pk), self)

    def cache_delete(self):
        self.CACHE.delete(self.cache_key(self.pk))
        self.cache_properties_delete()

    def cache_properties_delete(self):
        for prop in self.CACHED_PROPERTIES:
            if hasattr(self, prop):
                delattr(self, prop)

    def refresh_from_db(self, using=None, fields=None):
        super().refresh_from_db(using, fields)
        self.cache_delete()

    def load_cached_properties(self):
        for prop in self.CACHED_PROPERTIES:
            getattr(self, prop)


class CacheDependsMixin:
    CACHE_DEPENDS_FIELDS = []

    def _process_attr(self, attr, *rest):
        if isinstance(attr, Manager):
            for a in attr.all():
                self._process_attr(a, *rest)
        elif rest:
            self._cache_delete_depends(attr, *rest)
        elif attr:
            attr.cache_delete()

    def _cache_delete_depends(self, obj, field, *rest):
        attr = getattr(obj, field)
        self._process_attr(attr, *rest)

    def cache_delete_depends(self):
        for field in self.CACHE_DEPENDS_FIELDS:
            self._cache_delete_depends(self, *field.split('.'))


@receiver(post_save)
def post_save_cache(sender, instance, **kwargs):
    if issubclass(sender, CacheableMixin):
        instance.cache_delete()
    if issubclass(sender, CacheDependsMixin):
        instance.cache_delete_depends()


@receiver(pre_delete)
def pre_delete_cache(sender, instance, **kwargs):
    if issubclass(sender, CacheableMixin):
        instance.cache_delete()
    if issubclass(sender, CacheDependsMixin):
        instance.cache_delete_depends()



from .basic_api_errors import BasicAPIErrorTest
from .basic_api_test import BasicAPITest
from .test_api_doc import ApiDocTest
from .test_api_doc_search import TestAPIDocSearch
from .test_auth_api import AuthAPITest
from .test_body_mixin import TestBodyMixin
from .test_cache import TestCache
from .test_client import TestApiClient
from .test_fields_choices import TestFieldsChoices
from .test_fields_date import TestFieldsDate
from .test_fields_decimal import TestFieldsDecimal
from .test_fields_minutes_duration import TestFieldsMinutesDuration
from .test_fields_one2many import TestFieldsOne2Many
from .test_fields_rich_text import TestFieldsRichText
from .test_filters import FiltersTest
from .test_locale_api import LocaleAPITest
from .test_ordered_rest_view_mixin import TestOrderedRestViewMixin
from .test_perm_tests import PermTestsTest
from .test_permissions import TestPermissions
from .test_perms_api import PermsAPITest
from .test_post_migrate_groups_update import TestPostMigrateGroupsUpdate
from .test_profiling_middleware import ProfilingMiddlewareTest
from .test_settings import TestSettings
from .test_shortcuts import ShortcutTest
from .test_timezone_middleware import TimezoneMiddlewareTest
from .test_tools import TestTools
from .test_user_queryset import TestUserQuerySet
from .test_view_decorators import ViewDecoratorTest
from .test_view_fields import ViewFieldTest



from django.db.models import Q
from django.test import TestCase

from pfx.pfxcore.test import TestAssertMixin
from pfx.pfxcore.views import Filter


class MockParams:
    def __init__(self, **params):
        self.params = params

    def getlist(self, key):
        return self.params.get(key) or []


class FiltersTest(TestAssertMixin, TestCase):
    def test_filter_func(self):
        filter = Filter('test', "Test", filter_func=lambda v: Q(test=v))
        q = filter.query(MockParams(test=["A test value"]))
        self.assertEqual(str(q), "(AND: ('test', 'A test value'))")

    def test_filter_func_defaults(self):
        filter = Filter(
            'test', "Test", filter_func=lambda v: Q(test=v),
            defaults=["A default value"])
        q = filter.query(MockParams(test=[]))
        self.assertEqual(str(q), "(AND: ('test', 'A default value'))")
        q = filter.query(MockParams(test=["A test value"]))
        self.assertEqual(str(q), "(AND: ('test', 'A test value'))")

    def test_filter_func_multi(self):
        filter = Filter('test', "Test", filter_func=lambda v: Q(test=v))
        q = filter.query(MockParams(test=["A test value", "another"]))
        self.assertEqual(
            str(q), "(OR: ('test', 'A test value'), ('test', 'another'))")

    def test_filter_func_multi_and(self):
        filter = Filter(
            'test', "Test", filter_func=lambda v: Q(test=v),
            filter_func_and=True)
        q = filter.query(MockParams(test=["A test value", "another"]))
        self.assertEqual(
            str(q), "(AND: ('test', 'A test value'), ('test', 'another'))")

    def test_filter_func_multi_list(self):
        filter = Filter(
            'test', "Test", filter_func=lambda v: Q(test=v),
            filter_func_list=True)
        q = filter.query(MockParams(test=["A test value", "another"]))
        self.assertEqual(
            str(q), "(AND: ('test', ['A test value', 'another']))")



import logging
from functools import wraps

from django.utils.translation import gettext_lazy as _

from apispec.utils import deepupdate

from pfx.pfxcore.exceptions import APIError
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.settings import settings
from pfx.pfxcore.test import format_request

logger = logging.getLogger(__name__)


def rest_api(
        path, method='get', public=None, perms=None,
        priority=0, priority_doc=0, parameters=None,
        request_schema=None, response_schema=None, filters=False,
        search=False, groups=None):
    def decorator(func):
        @wraps(func)
        def wrapper(self, request, *args, **kwargs):
            self.request = request
            self.kwargs = kwargs
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("")
                logger.debug("##### REQUEST %s #####", request)
                logger.debug("")
            if (settings.PFX_TEST_MODE and
                    'HTTP_X_PRINT_REQUEST' in request.META):
                print(format_request(request))
            try:
                self.check_perm(public, func.__name__, perms, *args, **kwargs)
                return func(self, *args, **kwargs)
            except APIError as e:
                return e.response
            except Exception as e:
                logger.exception(e)
                return JsonResponse(dict(message=_(
                    "An internal server error occured.")), status=500)
        wrapper.rest_api_path = path
        wrapper.rest_api_method = method
        wrapper.rest_api_priority = priority
        wrapper.rest_api_priority_doc = priority_doc
        wrapper.rest_api_params = parameters or []
        wrapper.rest_api_request_schema = request_schema
        wrapper.rest_api_response_schema = response_schema
        wrapper.rest_api_filters = filters
        wrapper.rest_api_search = search
        wrapper.rest_api_groups = set(groups or [])
        wrapper.rest_api_public = public
        return wrapper
    return decorator


def rest_property(string=None, type="CharField", field=None, order=None):
    def decorator(func):
        func.short_description = string
        func.field_type = type
        func.field = field
        func.order = order
        return property(func)
    return decorator


def rest_view(path):
    def decorator(cls):
        cls.rest_view_path[cls] = path
        cls._rest_view_path = path
        return cls
    return decorator


def rest_doc(path, method, priority=None, **vals):
    def decorator(cls):
        if cls not in cls.rest_view_path:
            raise Exception(
                "@rest_doc must be used before a @rest_view decorator")
        key = f'{cls.rest_view_path[cls]}{path}', method
        cls.rest_doc[key] = deepupdate(cls.rest_doc.get(key, {}), vals)
        if priority is not None:
            cls.rest_doc_priority[key] = priority
        return cls
    return decorator



"""
Development settings

Load .dev_custom if it exists for personal settings (ignored by git).
Use .dev_default directly if no.

Use: `cp dev_custom_example.py dev_custom.py to create it.`
"""

try:
    from .dev_custom import *  # noqa
except ImportError:
    from .dev_default import *  # noqa



from django.contrib.auth.hashers import make_password
from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests_custom_user.models import User


class TestApi(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create(
            username='user',
            password=make_password('test'))

    def test_get_user(self):
        self.client.login(username='user')

        response = self.client.get(f'/api/users/{self.user.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'username', 'user')



import json
import logging
import re
from functools import reduce
from http.cookies import SimpleCookie
from json import JSONDecodeError

from django.db import transaction
from django.test.client import Client as DjangoClient

logger = logging.getLogger(__name__)


def format_response(r, title="http response"):
    res = []
    res.append(f"\n******************** {title} ********************")
    if not r:
        res.append("Response is null")
    res.append(f"Status: {r.status_code} {r.reason_phrase}")
    res.append("Headers: ")
    res.append("\n".join(f"  {k}: {v}" for k, v in r.headers.items()))
    res.append("Content: ")
    if hasattr(r, 'json_content'):
        res.append(json.dumps(r.json_content, indent=4, sort_keys=True))
    elif hasattr(r, 'content') and r.content:
        res.append(str(r.content))
    elif hasattr(r, 'streaming_content'):
        res.append("Streaming content")
    else:
        res.append("Response is empty")
    res.append("*******************************************************")
    return '\n'.join(res)


def format_request(r, title="http request"):
    res = []
    res.append(f"\n******************** {title} ********************")
    if not r:
        return "Request is null"
    res.append(f"Path: {r.path_info}")
    res.append(f"Method: {r.method}")
    res.append("Query params:")
    res.append("\n".join(f"  {k}: {v}" for k, v in r.GET.items()))
    res.append("Headers: ")
    res.append("\n".join(f"  {k}: {v}" for k, v in r.headers.items()))
    res.append("Content: ")
    if r.content_type == 'application/json':
        try:
            res.append(
                json.dumps(json.loads(r.body), indent=4, sort_keys=True))
        except JSONDecodeError:
            res.append("Json Decode Error")
            res.append(str(r.body))
    elif r.body:
        res.append(str(r.body))
    else:
        res.append("Request is empty")
    res.append("*******************************************************")
    return '\n'.join(res)


def get_auth_cookie(response):
    return [
        v for k, v in response.client.cookies.items()
        if k == 'token'][0]


def get_url(path, **params):
    if params:
        param_string = "&".join([f"{k}={v}" for k, v in params.items()])
        return f"{path}?{param_string}"
    return path


class APIClient(DjangoClient):

    token = None
    default_locale = None
    auth_cookie = None

    def __init__(self, enforce_csrf_checks=False, raise_request_exception=True,
                 default_locale=None, with_cookie=False, **defaults):
        super().__init__(enforce_csrf_checks, raise_request_exception,
                         **defaults)
        self.default_locale = default_locale
        self.with_cookie = with_cookie

    def login(self, username, password='test',
              path='/api/auth/login', locale=None, remember_me=False):
        self.token = None
        response = self.post(
            get_url(path, mode=(self.with_cookie and 'cookie' or 'jwt')),
            dict(
                username=username,
                password=password,
                remember_me=remember_me),
            locale=locale)
        if response.status_code == 200:
            if self.with_cookie:
                self.auth_cookie = get_auth_cookie(response)
                regex = r".*token=([\w\._-]*);.*"
                self.token = re.findall(regex, str(self.auth_cookie))[0]
            else:
                self.token = response.json_content['token']
            return True
        return False

    def logout(self):
        self.token = None

    def generic(self, method, path, data='',
                content_type='application/octet-stream', secure=False,
                locale=None, print_request=False, **extra):
        cookies = extra.pop('cookies', {})
        if self.token:
            if self.with_cookie:
                cookies['token'] = self.token
            else:
                extra.update(HTTP_AUTHORIZATION=f"Bearer {self.token}")
        if cookies:
            if self.cookies:
                for k, v in cookies.items():
                    self.cookies[k] = v
            else:
                self.cookies = SimpleCookie(cookies)
        if locale:
            extra.update(HTTP_X_CUSTOM_LANGUAGE=f"{locale}")
        elif self.default_locale:
            extra.update(HTTP_X_CUSTOM_LANGUAGE=f"{self.default_locale}")
        if print_request:
            extra.update(HTTP_X_PRINT_REQUEST="true")
        return super().generic(
            method, path, data, content_type, secure, **extra)

    @staticmethod
    def decode_response(response, method):
        if (response.headers['Content-Type'] == 'application/json'
                and method not in ['head', 'option', 'trace']):
            response.json_content = response.json()
        response.formatted = lambda: format_response(response)
        response.print = lambda: print(format_response(response))
        return response

    def get(self, path, data=None, secure=False, locale=None, **extra):
        return self.decode_response(
            super().get(
                path, data=data, secure=secure, locale=locale, **extra),
            method='get')

    def post(self, path, data=None, content_type='application/json',
             secure=False, locale=None, **extra):
        return self.decode_response(
            super().post(
                path, data=data, content_type=content_type, secure=secure,
                locale=locale, **extra), method='post')

    def head(self, path, data=None, secure=False, locale=None, **extra):
        return self.decode_response(
            super().head(
                path, data=data, secure=secure, locale=locale, **extra),
            method='head')

    def trace(self, path, secure=False, locale=None, **extra):
        return self.decode_response(super().trace(
            path, secure=secure, locale=locale, **extra), method='trace')

    def options(self, path, data='', content_type='application/json',
                secure=False, locale=None, **extra):
        return self.decode_response(
            super().options(
                path, data=data, content_type=content_type, secure=secure,
                locale=locale, **extra), method='options')

    def put(self, path, data='', content_type='application/json',
            secure=False, locale=None, **extra):
        return self.decode_response(
            super().put(
                path, data=data, content_type=content_type, secure=secure,
                locale=locale, **extra), method='put')

    def patch(self, path, data='', content_type='application/json',
              secure=False, locale=None, **extra):
        return self.decode_response(
            super().patch(
                path, data=data, content_type=content_type,
                secure=secure, locale=locale, **extra), method='patch')

    def delete(self, path, data='', content_type='application/json',
               secure=False, locale=None, **extra):
        return self.decode_response(
            super().delete(
                path, data=data, content_type=content_type, secure=secure,
                locale=locale, **extra), method='delete')


class TestAssertMixin:

    # assert response status code
    def assertRC(self, response, code, msg=None):
        msg = '\n'.join([msg or '', response.formatted()])
        return self.assertEqual(response.status_code, code, msg=msg)

    def format_json(self, src):
        if isinstance(src, dict):
            return json.dumps(src, indent=2)
        return src.formatted()

    def get_json(self, src):
        return src if isinstance(src, dict) else src.json_content

    def get_val(self, src, key, msg=None):
        def _p(k):
            return int(k[1:]) if k[0] == '@' else k

        try:
            return reduce(
                lambda c, k: c[_p(k)], key.split("."), self.get_json(src))
        except IndexError:
            assert False, f"'{key}' index not found in:" + msg
        except KeyError:
            assert False, f"'{key}' not found in:" + msg

    # assert JSON content at key equals value
    def assertJE(self, src, key, value, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        return self.assertEqual(self.get_val(
            src, key, msg=msg), value, msg=msg)

    # assert JSON content at key not equals value
    def assertNJE(self, src, key, value, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        return self.assertNotEqual(
            self.get_val(src, key, msg=msg), value, msg=msg)

    # assert JSON content at key exists
    def assertJEExists(self, src, key, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        if '.' not in key:
            return self.assertIn(key, self.get_json(src), msg=msg)
        path, name = key.rsplit('.', 1)
        return self.assertIn(name, self.get_val(src, path, msg=msg), msg=msg)

    # assert JSON content at key not exists
    def assertJENotExists(self, src, key, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        if '.' not in key:
            return self.assertNotIn(key, self.get_json(src), msg=msg)
        path, name = key.rsplit('.', 1)
        return self.assertNotIn(
            name, self.get_val(src, path, msg=msg), msg=msg)

    # assert JSON array size
    def assertSize(self, src, key, size, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        return self.assertEqual(
            len(self.get_val(src, key, msg=msg)), size, msg=msg)

    # assert JSON array contains
    def assertJIn(self, src, key, element, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        return self.assertIn(
            element, self.get_val(src, key, msg=msg), msg=msg)

    # assert JSON array contains
    def assertJNotIn(self, src, key, element, msg=None):
        msg = '\n'.join([msg or '', self.format_json(src)])
        return self.assertNotIn(
            element, self.get_val(src, key, msg=msg), msg=msg)


class TestPermsAssertMixin(TestAssertMixin):
    USER_TESTS = {}

    def assertPerms(self, **methods):
        ops = ['size']
        for method, result in methods.items():
            method_split = method.split('__')
            method = method_split.pop(0)
            op = (
                method_split and method_split[-1] in ops and
                method_split.pop(-1))
            test = '.'.join(method_split)
            with transaction.atomic():
                sid = transaction.savepoint()
                with self.subTest(msg=method):
                    response = getattr(self, method)()
                    msg = (
                        hasattr(self, '_logged_user') and
                        f"User {self._logged_user}" or None)
                    if test == 'count':
                        self.assertJE(response, 'meta.count', result, msg)
                        self.assertSize(response, 'items', result, msg)
                    elif test:
                        if op == 'size':
                            self.assertSize(response, test, result, msg)
                        else:
                            self.assertJE(response, test, result, msg)
                    else:
                        self.assertRC(response, result, msg)
                transaction.savepoint_rollback(sid)

    def test_all_users(self):
        for user, perms in self.USER_TESTS.items():
            self.client.login(username=user)
            self._logged_user = user
            self.assertPerms(**perms)


class MockBoto3Client:
    def generate_presigned_url(self, *args, **kwargs):
        params = kwargs.get('Params', {})
        return f"http://{params.get('Bucket')}/{params.get('Key')}"

    def head_object(self, *args, **kwargs):
        return dict(ContentLength=1000, ContentType="image/png")

    def delete_object(self, *args, **kwargs):
        pass



# from django.contrib.auth.models import AbstractUser

from .abstract_pfx_base_user import AbstractPFXUser


class PFXUser(AbstractPFXUser):
    """The Django User with PFX mixins.
    """

    class Meta(AbstractPFXUser.Meta):
        swappable = "AUTH_USER_MODEL"



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.utils.deprecation import MiddlewareMixin

import jwt
from jwt import DecodeError

from pfx.pfxcore.models import CacheableMixin
from pfx.pfxcore.shortcuts import delete_token_cookie

logger = logging.getLogger(__name__)


class JWTTokenDecodeMixin:

    @classmethod
    def get_cached_user(cls, pk):
        UserModel = get_user_model()
        has_cache = issubclass(UserModel, CacheableMixin)
        if (has_cache):
            user = UserModel.cache_get(pk)
            if user:
                return user
        user = UserModel._default_manager.get(pk=pk)
        if (has_cache):
            user.cache()
        return user

    @classmethod
    def decode_jwt_header(cls, token):
        try:
            headers = jwt.get_unverified_header(token)
            if 'pfx_user_pk' not in headers:
                raise jwt.InvalidTokenError(
                    "Missing pfx_user_pk in token headers")
            return headers['pfx_user_pk']
        except (jwt.ExpiredSignatureError,
                jwt.InvalidTokenError, jwt.InvalidSignatureError,
                DecodeError) as e:
            # Log these exceptions only in debug mode
            logger.debug(e, exc_info=True)
            raise
        except Exception as e:
            # Always logs unexpected exceptions
            logger.exception(e)
            raise

    @classmethod
    def decode_jwt(cls, token, otp_login=False):
        user_pk = cls.decode_jwt_header(token)
        try:
            user = cls.get_cached_user(user_pk)
            decoded = jwt.decode(
                token,
                user.get_user_jwt_signature_key() + settings.PFX_SECRET_KEY,
                options=dict(require=["exp"]),
                algorithms="HS256")
            if decoded.get('pfx_otp_login') and not otp_login:
                raise jwt.InvalidTokenError(
                    "This token is reserved for OTP login")
            return user, *decoded.get('pfx_login_options', ['jwt', False])
        except (get_user_model().DoesNotExist, jwt.ExpiredSignatureError,
                jwt.InvalidTokenError, jwt.InvalidSignatureError,
                DecodeError) as e:
            # Log these exceptions only in debug mode
            logger.debug(e, exc_info=True)
            raise
        except Exception as e:
            # Always logs unexpected exceptions
            logger.exception(e)
            raise


class AuthenticationMiddleware(JWTTokenDecodeMixin, MiddlewareMixin):
    """A middleware to authenticate with a bearer token.

    If `Authorization` is defined in request headers and the value is
    a valid JWT token (in the `"Bearer $TOKEN"` format), use it to
    set `request.user`. Otherwise, set it with
    `django.contrib.auth.models.AnonymousUser()`.
    """

    def process_request(self, request):
        authorization = request.headers.get('Authorization')
        if authorization:
            try:
                _, token = authorization.split("Bearer ")
            except ValueError:
                token = ""
            try:
                request.user, request.login_mode, request.login_remember_me = (
                    self.decode_jwt(token))
            except Exception:
                request.user = AnonymousUser()
        else:
            if not hasattr(request, 'user'):
                request.user = AnonymousUser()

    def process_response(self, request, response):
        return response


class CookieAuthenticationMiddleware(JWTTokenDecodeMixin, MiddlewareMixin):
    """A middleware to authenticate with a cookie.

    If `token` is defined in cookies and the value is
    a valid JWT token, use it to set `request.user`.
    Otherwise, set it with
    `django.contrib.auth.models.AnonymousUser()`.

    If the token is not valid, delete the cookie.

    If `request.delete_cookie` is `True` after the processing,
    delete the cookie.
    """

    def process_request(self, request):
        token = request.COOKIES.get('token', "")
        if token:
            try:
                request.user, request.login_mode, request.login_remember_me = (
                    self.decode_jwt(token))
            except Exception:
                request.user = AnonymousUser()
                request.delete_cookie = True
        else:
            if not hasattr(request, 'user'):
                request.user = AnonymousUser()

    def process_response(self, request, response):
        if getattr(request, 'delete_cookie', False):
            return delete_token_cookie(response)
        return response



from django.db import connection, models
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import include, path

from pfx.pfxcore import register_views
from pfx.pfxcore.decorator import rest_view
from pfx.pfxcore.models import OrderedModelMixin, PFXModelMixin
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import OrderedRestViewMixin, RestView
from tests.models import Author, User
from tests.views import FakeViewMixin


class OrderedModel(PFXModelMixin, OrderedModelMixin, models.Model):
    name = models.CharField("Name", max_length=30)

    class Meta:
        verbose_name = "OrderedModel"
        verbose_name_plural = "OrderedModels"
        ordering = ['seq', 'pk']

    def __str__(self):
        return self.name


class OrderedByNameModel(PFXModelMixin, OrderedModelMixin, models.Model):
    ordered_by = ['name']

    name = models.CharField("Name", max_length=30)

    class Meta:
        verbose_name = "OrderedByNameModel"
        verbose_name_plural = "OrderedByNameModels"
        ordering = ['name', 'seq', 'pk']

    def __str__(self):
        return f"{self.name} / {self.seq}"


class OrderedByNameFirstModel(PFXModelMixin, OrderedModelMixin, models.Model):
    ordered_default = 'first'
    ordered_by = ['name']

    name = models.CharField("Name", max_length=30)

    class Meta:
        verbose_name = "OrderedByNameModel"
        verbose_name_plural = "OrderedByNameModels"
        ordering = ['name', 'seq', 'pk']

    def __str__(self):
        return f"{self.name} / {self.seq}"


class OrderedByFkModel(PFXModelMixin, OrderedModelMixin, models.Model):
    ordered_by = ['ref']

    ref = models.ForeignKey(
        OrderedModel, verbose_name="OrderedModel",
        related_name='+', on_delete=models.RESTRICT)

    class Meta:
        verbose_name = "OrderedByFkModel"
        verbose_name_plural = "OrderedByFkModels"
        ordering = ['ref', 'seq', 'pk']

    def __str__(self):
        return f"{self.ref.name} / {self.seq}"


@rest_view("/ordered")
class OrderedModelRestView(FakeViewMixin, OrderedRestViewMixin, RestView):
    default_public = True
    model = OrderedModel
    fields = ['name', 'seq']


@rest_view("/ordered-by-name")
class OrderedByNameModelRestView(
        FakeViewMixin, OrderedRestViewMixin, RestView):
    default_public = True
    model = OrderedByNameModel
    fields = ['name', 'seq']


@rest_view("/ordered-by-fk")
class OrderedByFkModelRestView(FakeViewMixin, OrderedRestViewMixin, RestView):
    default_public = True
    model = OrderedByFkModel
    fields = ['ref', 'seq']


@rest_view("/ordered-authors")
class OrderedAuthorRestView(FakeViewMixin, OrderedRestViewMixin, RestView):
    default_public = True
    model = Author


class conf:
    urlpatterns = [
        path('api/', include(register_views(
            OrderedModelRestView,
            OrderedByNameModelRestView,
            OrderedByFkModelRestView,
        ))),
        path('api/', include('pfx.pfxcore.urls'))
    ]


class conf_failed:
    urlpatterns = [
        path('api/', include(register_views(
            OrderedAuthorRestView,
        ))),
        path('api/', include('pfx.pfxcore.urls'))
    ]


def order(model):
    return list(model.objects.all().values_list('pk', flat=True))


@override_settings(ROOT_URLCONF=conf)
class TestOrderedRestViewMixin(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(OrderedModel)
            schema_editor.create_model(OrderedByNameModel)
            schema_editor.create_model(OrderedByNameFirstModel)
            schema_editor.create_model(OrderedByFkModel)
        cls.user = User.objects.create_user(
            username='jrr.tolkien',
            email="jrr.tolkien@oxford.com",
            password='RIGHT PASSWORD',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
        )
        cls.o1 = OrderedModel.objects.create(name='A')
        cls.o2 = OrderedModel.objects.create(name='A')
        cls.o3 = OrderedModel.objects.create(name='A')
        cls.obn1 = OrderedByNameModel.objects.create(name='AAA')
        cls.obn2 = OrderedByNameModel.objects.create(name='BBB')
        cls.obn3 = OrderedByNameModel.objects.create(name='BBB')
        cls.obnf1 = OrderedByNameFirstModel.objects.create(name='AAA')
        cls.obnf2 = OrderedByNameFirstModel.objects.create(name='BBB')
        cls.obnf3 = OrderedByNameFirstModel.objects.create(name='BBB')
        cls.obfk1 = OrderedByFkModel.objects.create(ref=cls.o1)
        cls.obfk2 = OrderedByFkModel.objects.create(ref=cls.o2)
        cls.obfk3 = OrderedByFkModel.objects.create(ref=cls.o2)

    def test_default_values(self):
        response = self.client.get('/api/ordered?items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        self.assertJE(response, 'items.@0.pk', self.o1.pk)
        self.assertJE(response, 'items.@1.pk', self.o2.pk)
        self.assertJE(response, 'items.@2.pk', self.o3.pk)
        self.assertJE(response, 'items.@0.seq', 1)
        self.assertJE(response, 'items.@1.seq', 2)
        self.assertJE(response, 'items.@2.seq', 3)

    def test_default_values_by_name(self):
        response = self.client.get('/api/ordered-by-name?items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        self.assertJE(response, 'items.@0.pk', self.obn1.pk)
        self.assertJE(response, 'items.@1.pk', self.obn2.pk)
        self.assertJE(response, 'items.@2.pk', self.obn3.pk)
        self.assertJE(response, 'items.@0.seq', 1)
        self.assertJE(response, 'items.@1.seq', 1)
        self.assertJE(response, 'items.@2.seq', 2)

    def test_default_values_by_fk(self):
        response = self.client.get('/api/ordered-by-fk?items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        self.assertJE(response, 'items.@0.pk', self.obfk1.pk)
        self.assertJE(response, 'items.@1.pk', self.obfk2.pk)
        self.assertJE(response, 'items.@2.pk', self.obfk3.pk)
        self.assertJE(response, 'items.@0.seq', 1)
        self.assertJE(response, 'items.@1.seq', 1)
        self.assertJE(response, 'items.@2.seq', 2)

    def test_up(self):
        response = self.client.put(
            f'/api/ordered/{self.o3.pk}?move=up', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o1.pk, self.o3.pk, self.o2.pk])

        # Check do nothings if first
        response = self.client.put(
            f'/api/ordered/{self.o1.pk}?move=up', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o1.pk, self.o3.pk, self.o2.pk])

    def test_up_by_name(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}?move=up', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

        # Check do nothings if first
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}?move=up', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

    def test_up_by_fk(self):
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk3.pk}?move=up', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

        # Check do nothings if first
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk3.pk}?move=up', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

    def test_down(self):
        response = self.client.put(
            f'/api/ordered/{self.o2.pk}?move=down', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o1.pk, self.o3.pk, self.o2.pk])

        # Check do nothings if last
        response = self.client.put(
            f'/api/ordered/{self.o2.pk}?move=down', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o1.pk, self.o3.pk, self.o2.pk])

    def test_down_by_name(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn2.pk}?move=down', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

        # Check do nothings if last
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn2.pk}?move=down', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

    def test_down_by_fk(self):
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk2.pk}?move=down', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

        # Check do nothings if last
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk2.pk}?move=down', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

    def test_top(self):
        response = self.client.put(
            f'/api/ordered/{self.o3.pk}?move=top', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o3.pk, self.o1.pk, self.o2.pk])

        # Check do nothings if first
        response = self.client.put(
            f'/api/ordered/{self.o3.pk}?move=top', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o3.pk, self.o1.pk, self.o2.pk])

    def test_top_by_name(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}?move=top', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

        # Check do nothings if first
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}?move=top', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

    def test_top_by_fk(self):
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk3.pk}?move=top', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

        # Check do nothings if first
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk3.pk}?move=top', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

    def test_bottom(self):
        response = self.client.put(
            f'/api/ordered/{self.o1.pk}?move=bottom', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o2.pk, self.o3.pk, self.o1.pk])

        # Check do nothings if last
        response = self.client.put(
            f'/api/ordered/{self.o1.pk}?move=bottom', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o2.pk, self.o3.pk, self.o1.pk])

    def test_bottom_by_name(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn2.pk}?move=bottom', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

        # Check do nothings if last
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn2.pk}?move=bottom', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

    def test_bottom_by_fk(self):
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk2.pk}?move=bottom', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

        # Check do nothings if last
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk2.pk}?move=bottom', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

    def test_above(self):
        response = self.client.put(
            f'/api/ordered/{self.o3.pk}?move=above&object={self.o2.pk}', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o1.pk, self.o3.pk, self.o2.pk])

    def test_above_by_name(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}'
            f'?move=above&object={self.obn2.pk}', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

    def test_above_by_fk(self):
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk3.pk}'
            f'?move=above&object={self.obfk2.pk}', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

    def test_below(self):
        response = self.client.put(
            f'/api/ordered/{self.o1.pk}?move=below&object={self.o2.pk}', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedModel),
            [self.o2.pk, self.o1.pk, self.o3.pk])

    def test_below_by_name(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn2.pk}'
            f'?move=below&object={self.obn3.pk}', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByNameModel),
            [self.obn1.pk, self.obn3.pk, self.obn2.pk])

    def test_below_by_fk(self):
        response = self.client.put(
            f'/api/ordered-by-fk/{self.obfk2.pk}'
            f'?move=below&object={self.obfk3.pk}', {})
        self.assertRC(response, 200)

        self.assertEqual(
            order(OrderedByFkModel),
            [self.obfk1.pk, self.obfk3.pk, self.obfk2.pk])

    @override_settings(ROOT_URLCONF=conf_failed)
    def test_no_ordered_mode_mixin_view(self):
        with self.assertRaises(TypeError):
            self.client.get('/api/ordered-authors')

    def test_missing_object_param(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}?move=above', {})
        self.assertRC(response, 400)

    def test_object_other_order_parent(self):
        response = self.client.put(
            f'/api/ordered-by-name/{self.obn3.pk}'
            f'?move=above&object={self.obn1.pk}', {})
        self.assertRC(response, 400)

    def test_create_default(self):
        response = self.client.post(
            '/api/ordered-by-name', dict(name='BBB'))
        self.assertRC(response, 200)
        self.assertJE(response, 'seq', 3)

    def test_model_ordered_default(self):
        self.obn3.name = self.obn1.name
        self.obn3.save()
        self.obn3.refresh_from_db()
        self.assertEqual(self.obn3.seq, 2)

        self.obnf3.name = self.obnf1.name
        self.obnf3.save()
        self.obnf3.refresh_from_db()
        self.assertEqual(self.obnf3.seq, 1)



import hashlib
import logging
from pathlib import Path

from django.http import FileResponse

from pfx.pfxcore.shortcuts import settings

logger = logging.getLogger(__name__)


class LocalStorage:
    direct = True

    def to_python(self, value):
        return value

    def get_url(self, request, key):
        return Path(settings.STORAGE_LOCAL_ROOT, key)

    def upload(self, key, file, **kwargs):
        key_list = key.split('/')
        filename = key_list[-1]
        relative_path = Path(*key_list[:-1])
        dirname = Path(settings.STORAGE_LOCAL_ROOT, relative_path)
        dirname.mkdir(parents=True, exist_ok=True)
        suffixes = Path(filename).suffixes
        if not suffixes:
            ext = ''
        elif suffixes[-1] in ['gz', 'bz2'] and len(suffixes) > 1:
            ext = ''.join(suffixes[-2:])
        else:
            ext = suffixes[-1]
        hashname = f'{hashlib.sha1(file).hexdigest()}{ext}'
        final_key = str(Path(relative_path, hashname))
        with open(Path(dirname, hashname), 'wb') as f:
            f.write(file)
        with open(Path(dirname, hashname), 'rb') as f:
            response = FileResponse(
                f, as_attachment=True, filename=filename)
            return {
                'key': final_key,
                'name': filename,
                'content-length': response.get('Content-Length'),
                'content-type': response.get('Content-Type'),
            }

    def delete(self, value):
        if not value or 'key' not in value:
            return value  # pragma: no cover
        path = Path(settings.STORAGE_LOCAL_ROOT, value['key'])
        path.unlink(missing_ok=True)
        while True:
            path = path.parent
            if path == Path(settings.STORAGE_LOCAL_ROOT):
                break
            if not path.exists():
                break
            if any(path.iterdir()):
                break
            try:
                path.rmdir()
            except FileNotFoundError:
                pass



import logging

from pfx.pfxcore.decorator import rest_view
from pfx.pfxcore.views import RestView
from tests_custom_user.models import User

logger = logging.getLogger(__name__)


@rest_view("/users")
class UserRestView(RestView):
    model = User
    fields = ['username']



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class SubsetPageSubset(QueryParameter):
    """The maximum number of pages in `pagination` mode.

    The maximum numb of pages in `meta.subset.subset` element.

    If the value of `subset` is not `"pagination"`, this attribute will
    be ignored.
    """
    name = 'page_subset'
    schema = dict(type='integer', default=5)


register_global_parameter(SubsetPageSubset)



from django.core.exceptions import NON_FIELD_ERRORS
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.http import JsonResponse

from .shortcuts import delete_token_cookie, f


def validation_error_dict(e):
    if hasattr(e, 'error_dict'):
        return dict(e)
    return {NON_FIELD_ERRORS: list(e)}


class APIError(Exception):
    def __init__(
            self, message, status=400, delete_cookie=False,
            validation_error=None, **kwargs):
        self.status = status
        if validation_error:
            d = validation_error_dict(validation_error)
            mlist = d.get(NON_FIELD_ERRORS)
            if mlist:
                message = mlist[0]
            self.data = dict(message=message, errors=d, **kwargs)
        else:
            self.data = dict(message=message, **kwargs)
        self.delete_cookie = delete_cookie

    @property
    def response(self):
        res = JsonResponse(self.data, status=self.status)
        if self.delete_cookie:
            return delete_token_cookie(res)
        return res


class ModelNotFoundAPIError(APIError):
    def __init__(self, model, status=404, **kwargs):
        super().__init__(
            f(_("{model} not found."), model=model._meta.verbose_name),
            status=status, **kwargs)


class JsonErrorAPIError(APIError):
    def __init__(self, json_error, status=422, **kwargs):
        super().__init__(
            f(_("JSON Malformed {}").format(str(json_error))),
            status=status, **kwargs)


class ModelValidationAPIError(APIError):
    def __init__(
            self, errors, status=422, delete_cookie=False, **kwargs):
        self.status = status
        self.data = errors
        self.delete_cookie = delete_cookie


class RelatedModelNotFoundAPIError(ModelValidationAPIError):
    def __init__(
            self, field, model, status=422, delete_cookie=False, **kwargs):
        super().__init__({field: [
            f(_("{model} not found."), model=model._meta.verbose_name)]},
            status=status, delete_cookie=delete_cookie, **kwargs)


class AuthenticationError(APIError):
    def __init__(self, message=None, status=422, **kwargs):
        super().__init__(
            f(message or _("Login failed")),
            status=status, **kwargs)


class UnauthorizedError(APIError):
    def __init__(self, message=None, status=401, **kwargs):
        super().__init__(
            f(message or _("Unauthorized")),
            status=status, **kwargs)


class ForbiddenError(APIError):
    def __init__(self, message=None, status=403, **kwargs):
        super().__init__(
            f(message or _("Forbidden")),
            status=status, **kwargs)


class NotFoundError(APIError):
    def __init__(self, message=None, status=404, **kwargs):
        super().__init__(
            f(message or _("Resource not found")),
            status=status, **kwargs)



from datetime import date

from django.db import connection, models
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import include, path

from pfx.pfxcore import register_views
from pfx.pfxcore.decorator import rest_view
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import RestView
from tests.views import FakeViewMixin


class TestDateModel(JSONReprMixin, models.Model):
    date_at = models.DateField(blank=True, null=True)

    class Meta:
        verbose_name = "TestModel"
        verbose_name_plural = "TestModels"
        ordering = ['pk']


@rest_view("/test")
class TestDateRestView(FakeViewMixin, RestView):
    default_public = True
    model = TestDateModel


urlpatterns = [
    path('api/', include(register_views(TestDateRestView))),
    path('api/', include('pfx.pfxcore.urls'))
]


@override_settings(ROOT_URLCONF=__name__)
class TestFieldsDate(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestDateModel)

    def test_update_null(self):
        c1 = TestDateModel.objects.create(date_at=date(2025, 5, 3))

        response = self.client.put(f'/api/test/{c1.pk}', dict(date_at=None))
        self.assertRC(response, 200)
        self.assertJE(response, 'date_at', None)

    def test_update_empty_string(self):
        c1 = TestDateModel.objects.create(date_at=date(2025, 5, 3))

        response = self.client.put(f'/api/test/{c1.pk}', dict(date_at=""))
        self.assertRC(response, 200)
        self.assertJE(response, 'date_at', None)



from .abstract_pfx_base_user import AbstractPFXBaseUser, AbstractPFXUser
from .cache_mixins import CacheableMixin, CacheDependsMixin
from .login_ban import LoginBan
from .not_null_fields import (
    NotNullCharField,
    NotNullTextField,
    NotNullURLField,
)
from .ordered_model_mixin import OrderedModelMixin
from .otp_user_mixin import OtpUserMixin
from .pfx_models import (
    ErrorMessageMixin,
    JSONReprMixin,
    PFXModelMixin,
    UniqueConstraint,
)
from .pfx_user import PFXUser
from .user_filtered_queryset_mixin import UserFilteredQuerySetMixin



from django.test import TestCase

from pfx.pfxcore.test import TestAssertMixin
from tests.models import Author, User


class TestUserQuerySet(TestAssertMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(
            username='user',
            email="user@example.com",
            password='test',
            first_name='User',
            last_name='Test')
        cls.admin = User.objects.create_user(
            username='admin',
            email="admin@example.com",
            password='test',
            first_name='Admin',
            last_name='Test',
            is_superuser=True)

        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        cls.author2 = Author.objects.create(
            first_name='Philip Kindred',
            last_name='Dick',
            science_fiction=True,
            slug='philip-k-dick')

    def test_user_queryset(self):
        authors = Author._default_manager.all()
        self.assertEqual(len(authors), 2)
        authors = Author.user_objects.user(self.user)
        self.assertEqual(len(authors), 1)
        authors = Author.user_objects.user(self.admin)
        self.assertEqual(len(authors), 2)

    def test_bad_user_queryset(self):
        with self.assertRaises(NotImplementedError):
            Author.bad_user_objects.user(self.user)



from datetime import timedelta

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.settings import settings


class LoginBanQuerySet(models.QuerySet):
    def is_ban(self, username):
        if not username or settings.PFX_LOGIN_BAN_FAILED_NUMBER == 0:
            return False
        try:
            ban = self.get(username=username)
        except LoginBan.DoesNotExist:
            return False
        if ban.failed_counter % settings.PFX_LOGIN_BAN_FAILED_NUMBER == 0:
            seconds = settings.PFX_LOGIN_BAN_SECONDS_START + (
                settings.PFX_LOGIN_BAN_SECONDS_STEP * (
                    ban.failed_counter //
                    settings.PFX_LOGIN_BAN_FAILED_NUMBER - 1))
            ban_time = ban.last_failed + timedelta(seconds=seconds)
            now = timezone.now()
            if now < ban_time:
                return ban_time - now
        return False

    def ban(self, username):
        if not username:
            return
        try:
            ban = self.get(username=username)
            ban.save()
        except LoginBan.DoesNotExist:
            LoginBan.objects.create(username=username)

    def unban(self, username):
        if not username:
            return
        self.filter(username=username).delete()


class LoginBan(models.Model):
    username = models.CharField(_("Username"), max_length=150, unique=True)
    failed_counter = models.IntegerField(_("Failed counter"))
    last_failed = models.DateTimeField(_("Last failed"), auto_now=True)

    objects = LoginBanQuerySet.as_manager()

    class Meta:
        verbose_name = _("Login ban")
        verbose_name_plural = _("Login bans")

    def __str__(self):
        return self.username

    def save(self, *args, **kwargs):
        self.failed_counter = (self.failed_counter or 0) + 1
        return super().save(*args, **kwargs)



from django.urls import include, path

from pfx.pfxcore import register_views

from . import views

handler404 = 'pfx.pfxcore.views.resource_not_found'

urlpatterns = [
    path('api/', include(register_views(
        views.AuthorRestView,
        views.AuthorExtraMetaRestView,
        views.AuthorAnnotateRestView,
        views.AuthorFieldsPropsRestView,
        views.PrivateEditAuthorRestView,
        views.PrivateAuthorRestView,
        views.PermsAuthorRestView,
        views.AdminEditAuthorRestView,
        views.AdminAuthorRestView,
        views.BookRestView,
        views.BookCustomAuthorRestView,
        views.PermsBookRestView,
        views.BookTypeRestView,
        views.Testi18nView,
        views.TestErrorView,
        views.TestTimezoneView))),
    path('api/does-not-exists', views.AuthorRestView.as_view(
        pfx_methods=dict(get='does_not_exists'))),
    path('api/', include('pfx.pfxcore.urls'))
]



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class ListSearch(QueryParameter):
    """Filter result by a text search.

    Apply the default text search (object-dependent).
    """
    name = 'search'
    schema = dict(type='string', default=None)


register_global_parameter(ListSearch)



import logging
from importlib import import_module
from urllib.request import urlopen

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_delete, post_init
from django.dispatch import receiver

from pfx.pfxcore.shortcuts import settings

logger = logging.getLogger(__name__)

DEFAULT_KEYS = {"key", "name", "content-type", "content-length"}


def get_storage_class(class_path):
    ps = class_path.split('.')
    return getattr(import_module('.'.join(ps[:-1])), ps[-1])()


def validate_media_json(value):
    if isinstance(value, dict):
        missing_keys = []
        if value.get('key'):
            # Empty dict or empty key is allowed to save
            # and object without file to have a pk to determine the key.
            # When the key is set, other required keys must be present.
            if not value.get('name'):
                missing_keys.append('name')
            if not value.get('content-type'):
                missing_keys.append('content-type')
            if not value.get('content-length'):
                missing_keys.append('content-length')
        if missing_keys:
            raise ValidationError(
                f"Missing required keys: {', '.join(missing_keys)}")
    else:
        raise ValidationError("Value must be a dictionary.")


class MediaField(models.JSONField):
    def __init__(
            self, *args, max_length=255, get_key=None, storage=None,
            auto_delete=False, extra_keys=None, **kwargs):
        self.get_key = get_key or self.get_default_key
        if not storage and not settings.STORAGE_DEFAULT:
            raise Exception(
                "Missing storage. You have to set a storage "
                "class on the field or define STORAGE_DEFAULT settings.")
        self.storage = storage or get_storage_class(settings.STORAGE_DEFAULT)
        self.auto_delete = auto_delete
        self.allowed_keys = DEFAULT_KEYS.union(extra_keys or set())
        self._db_value = None

        # Inject default validator if empty
        kwargs.setdefault("validators", [validate_media_json])

        super().__init__(*args, max_length=max_length, **kwargs)

    @staticmethod
    def get_default_key(obj, filename):
        return f"{type(obj).__name__}/{obj.pk}/{filename}"

    def clean_value(self, value):
        if isinstance(value, dict):
            return {k: v for k, v in value.items() if k in self.allowed_keys}
        return value

    def get_prep_value(self, value):
        value = self.clean_value(value)
        return super().get_prep_value(value)

    def media_pre_save(self, obj):
        def save_file(obj, file, name):
            setattr(obj, self.name, self.upload(obj, file, name))
            obj.save(update_fields={self.name})

        after_save = []
        value = self.value_from_object(obj)
        if isinstance(value, dict):
            b64 = value.pop('base64', None)
            name = value.get('name')
        else:
            b64 = None

        if self._db_value and (b64 or not value):
            after_save.append(lambda: self.delete(self._db_value))
            setattr(obj, self.name, None)
        if b64:
            with urlopen(b64) as response:
                file = response.read()
            setattr(obj, self.name, dict(key=self.get_key(obj, file)))
            after_save.append(lambda: save_file(obj, file, name))
        return after_save

    def to_python(self, value):
        return super().to_python(self.storage.to_python(value))

    def get_upload_url(self, request, obj, filename):
        key = self.get_key(obj, filename)
        url = self.storage.get_upload_url(request, key)
        return dict(url=url, file=dict(name=filename, key=key))

    def get_url(self, request, obj):
        return self.storage.get_url(
            request, self.value_from_object(obj)['key'])

    def upload(self, obj, file, filename, **kwargs):
        key = self.get_key(obj, filename)
        return self.to_python(self.storage.upload(key, file, **kwargs))

    def delete(self, value):
        return self.storage.delete(value)


@receiver(post_init)
def post_init_media(sender, instance, **kwargs):
    for field in sender._meta.fields:
        if isinstance(field, MediaField):
            field._db_value = field.value_from_object(instance)


@receiver(post_delete)
def post_delete_media(sender, instance, **kwargs):
    for field in sender._meta.fields:
        if isinstance(field, MediaField) and field.auto_delete:
            field.storage.delete(field.value_from_object(instance))
